@@ -11,14 +11,12 @@ OC.L10N.register(
     "New folder" : "New folder",
     "Search" : "Search",
     "Empty trash bin" : "Empty trash bin",
-    "_%n folder_::_%n folders_" : ["%n folder","%n folders"],
     "Submit" : "Submit",
     "Cancel" : "Cancel",
     "Recent" : "Recent",
     "Shared with you" : "Shared with you",
     "Files" : "ფაილები",
     "Delete" : "Delete",
-    "Settings" : "Settings",
     "Bookmarklet" : "Bookmarklet",
     "Need help?" : "Need help?",
     "Link" : "Link",
