@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2020. The Nextcloud Bookmarks contributors.
+ * Copyright (c) 2020-2024. The Nextcloud Bookmarks contributors.
  *
  * This file is licensed under the Affero General Public License version 3 or later. See the COPYING file.
  */
@@ -7,23 +7,28 @@
 import axios from '@nextcloud/axios'
 import { generateUrl } from '@nextcloud/router'
 import { loadState } from '@nextcloud/initial-state'
-import AppGlobal from '../mixins/AppGlobal'
-import { mutations } from './mutations'
+import AppGlobal from '../mixins/AppGlobal.js'
+import { mutations } from './mutations.js'
 import * as Parallel from 'async-parallel'
-import uniq from 'lodash/uniq'
-import difference from 'lodash/difference'
+import uniq from 'lodash/uniq.js'
+import difference from 'lodash/difference.js'
 
-const BATCH_SIZE = 42
+const BATCH_SIZE = 100
 
 export const actions = {
 	ADD_ALL_BOOKMARKS: 'ADD_ALL_BOOKMARKS',
 	COUNT_BOOKMARKS: 'COUNT_BOOKMARKS',
 	COUNT_UNAVAILABLE: 'COUNT_UNAVAILABLE',
 	COUNT_ARCHIVED: 'COUNT_ARCHIVED',
+	COUNT_DUPLICATED: 'COUNT_DUPLICATED',
+	COUNT_ALL_CLICKS: 'COUNT_ALL_CLICKS',
+	COUNT_WITH_CLICKS: 'COUNT_WITH_CLICKS',
+
 	CREATE_BOOKMARK: 'CREATE_BOOKMARK',
 	FIND_BOOKMARK: 'FIND_BOOKMARK',
 	LOAD_BOOKMARK: 'LOAD_BOOKMARK',
 	DELETE_BOOKMARK: 'DELETE_BOOKMARK',
+	UNDELETE_BOOKMARK: 'UNDELETE_BOOKMARK',
 	OPEN_BOOKMARK: 'OPEN_BOOKMARK',
 	SAVE_BOOKMARK: 'SAVE_BOOKMARK',
 	MOVE_BOOKMARK: 'MOVE_BOOKMARK',
@@ -31,15 +36,19 @@ export const actions = {
 	CLICK_BOOKMARK: 'CLICK_BOOKMARK',
 	IMPORT_BOOKMARKS: 'IMPORT_BOOKMARKS',
 	DELETE_BOOKMARKS: 'DELETE_BOOKMARKS',
+	LOAD_DELETED_BOOKMARKS: 'LOAD_DELETED_BOOKMARKS',
 
 	LOAD_TAGS: 'LOAD_TAGS',
 	RENAME_TAG: 'RENAME_TAG',
 	DELETE_TAG: 'DELETE_TAG',
 
 	LOAD_FOLDERS: 'LOAD_FOLDERS',
+	LOAD_DELETED_FOLDERS: 'LOAD_DELETED_FOLDERS',
 	CREATE_FOLDER: 'CREATE_FOLDER',
 	SAVE_FOLDER: 'SAVE_FOLDER',
+	MOVE_FOLDER: 'MOVE_FOLDER',
 	DELETE_FOLDER: 'DELETE_FOLDER',
+	UNDELETE_FOLDER: 'UNDELETE_FOLDER',
 	LOAD_FOLDER_CHILDREN_ORDER: 'LOAD_FOLDER_CHILDREN_ORDER',
 	OPEN_FOLDER_DETAILS: 'OPEN_FOLDER_DETAILS',
 	OPEN_FOLDER_SHARING: 'OPEN_FOLDER_SHARING',
@@ -57,16 +66,19 @@ export const actions = {
 	FILTER_BY_UNTAGGED: 'FILTER_BY_UNTAGGED',
 	FILTER_BY_UNAVAILABLE: 'FILTER_BY_UNAVAILABLE',
 	FILTER_BY_ARCHIVED: 'FILTER_BY_ARCHIVED',
+	FILTER_BY_DUPLICATED: 'FILTER_BY_DUPLICATED',
 	FILTER_BY_TAGS: 'FILTER_BY_TAGS',
 	FILTER_BY_FOLDER: 'FILTER_BY_FOLDER',
+	FILTER_BY_SHARED_FOLDERS: 'FILTER_BY_SHARED_FOLDERS',
 	FILTER_BY_SEARCH: 'FILTER_BY_SEARCH',
 	FETCH_PAGE: 'FETCH_PAGE',
 	FETCH_ALL: 'FETCH_ALL',
 
 	SET_SETTING: 'SET_SETTING',
 	LOAD_SETTING: 'LOAD_SETTING',
-	LOAD_SETTINGS: 'SLOAD_SETTINGS',
+	LOAD_SETTINGS: 'LOAD_SETTINGS',
 
+	LOAD_SHARES: 'LOAD_SHARES',
 	LOAD_SHARES_OF_FOLDER: 'LOAD_SHARES_OF_FOLDER',
 	CREATE_SHARE: 'CREATE_SHARE',
 	EDIT_SHARE: 'EDIT_SHARE',
@@ -75,7 +87,9 @@ export const actions = {
 	LOAD_PUBLIC_LINK: 'LOAD_PUBLIC_LINK',
 	CREATE_PUBLIC_LINK: 'CREATE_PUBLIC_LINK',
 	DELETE_PUBLIC_LINK: 'DELETE_PUBLIC_LINK',
+	LOAD_SHARED_FOLDERS: 'LOAD_SHARED_FOLDERS',
 
+	EMPTY_TRASHBIN: 'EMPTY_TRASHBIN',
 }
 
 export default {
@@ -85,10 +99,20 @@ export default {
 		}
 	},
 
-	async [actions.COUNT_UNAVAILABLE]({ commit, dispatch, state }, folderId) {
+	async [actions.COUNT_UNAVAILABLE]({ commit, dispatch, state }) {
+		if (state.unavailableCount === null) {
+			try {
+				const count = loadState('bookmarks', 'unavailableCount')
+				return commit(mutations.SET_UNAVAILABLE_COUNT, count)
+			} catch (e) {
+				console.warn(
+					'Could not load initial unavailable bookmarks count state, continuing with HTTP request',
+				)
+			}
+		}
 		try {
 			const response = await axios.get(
-				url(state, '/bookmark/unavailable')
+				url(state, '/bookmark/unavailable'),
 			)
 			const {
 				data: { item: count, data, status },
@@ -103,13 +127,23 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to count unavailable bookmarks'
-				)
+					'Failed to count unavailable bookmarks',
+				),
 			)
 			throw err
 		}
 	},
-	async [actions.COUNT_ARCHIVED]({ commit, dispatch, state }, folderId) {
+	async [actions.COUNT_ARCHIVED]({ commit, dispatch, state }) {
+		if (state.archivedCount === null) {
+			try {
+				const count = loadState('bookmarks', 'archivedCount')
+				return commit(mutations.SET_ARCHIVED_COUNT, count)
+			} catch (e) {
+				console.warn(
+					'Could not load initial archived bookmarks count state, continuing with HTTP request',
+				)
+			}
+		}
 		try {
 			const response = await axios.get(url(state, '/bookmark/archived'))
 			const {
@@ -125,16 +159,79 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to count archived bookmarks'
-				)
+					'Failed to count archived bookmarks',
+				),
 			)
 			throw err
 		}
 	},
+	async [actions.COUNT_DUPLICATED]({ commit, dispatch, state }) {
+		if (state.duplicatedCount === null) {
+			try {
+				const count = loadState('bookmarks', 'duplicatedCount')
+				return commit(mutations.SET_DUPLICATED_COUNT, count)
+			} catch (e) {
+				console.warn(
+					'Could not load initial duplicated bookmarks count state, continuing with HTTP request',
+				)
+			}
+		}
+		try {
+			const response = await axios.get(url(state, '/bookmark/duplicated'))
+			const {
+				data: { item: count, data, status },
+			} = response
+			if (status !== 'success') {
+				throw new Error(data)
+			}
+			commit(mutations.SET_DUPLICATED_COUNT, count)
+		} catch (err) {
+			console.error(err)
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t(
+					'bookmarks',
+					'Failed to count duplicated bookmarks',
+				),
+			)
+			throw err
+		}
+	},
+	async [actions.COUNT_ALL_CLICKS]({ commit, dispatch, state }) {
+		try {
+			const count = loadState('bookmarks', 'allClicksCount')
+			return commit(mutations.SET_ALL_CLICKS_COUNT, count)
+		} catch (e) {
+			console.warn(
+				'Could not load initial all clicks count state',
+			)
+		}
+	},
+	async [actions.COUNT_WITH_CLICKS]({ commit, dispatch, state }) {
+		try {
+			const count = loadState('bookmarks', 'withClicksCount')
+			return commit(mutations.SET_WITH_CLICKS_COUNT, count)
+		} catch (e) {
+			console.warn(
+				'Could not load initial bookmarks with clicks count state',
+			)
+		}
+	},
 	async [actions.COUNT_BOOKMARKS]({ commit, dispatch, state }, folderId) {
+		if (String(folderId) === String(-1) && typeof state.countsByFolder[folderId] === 'undefined') {
+			try {
+				const count = loadState('bookmarks', 'allCount')
+				return commit(mutations.SET_BOOKMARK_COUNT, { folderId, count })
+			} catch (e) {
+				console.warn(e)
+				console.warn(
+					'Could not load initial bookmarks count state, continuing with HTTP request',
+				)
+			}
+		}
 		try {
 			const response = await axios.get(
-				url(state, `/folder/${folderId}/count`)
+				url(state, `/folder/${folderId}/count`),
 			)
 			const {
 				data: { item: count, data, status },
@@ -147,7 +244,7 @@ export default {
 			console.error(err)
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to count bookmarks')
+				AppGlobal.methods.t('bookmarks', 'Failed to count bookmarks'),
 			)
 			throw err
 		}
@@ -167,7 +264,7 @@ export default {
 			console.error(err)
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to load bookmark')
+				AppGlobal.methods.t('bookmarks', 'Failed to load bookmark'),
 			)
 			throw err
 		}
@@ -195,8 +292,8 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to find existing bookmark'
-				)
+					'Failed to find existing bookmark',
+				),
 			)
 			throw err
 		}
@@ -278,17 +375,13 @@ export default {
 				})
 				dispatch(actions.LOAD_FOLDER_CHILDREN_ORDER, -1)
 			}
-			dispatch(actions.RELOAD_VIEW)
-
-			// open sidebar
-			return dispatch(actions.OPEN_BOOKMARK, bookmark.id)
 		} catch (err) {
 			console.error(err)
 			commit(mutations.FETCH_END, 'createBookmark')
 			commit(mutations.REMOVE_BOOKMARK, prelimBookmark.id)
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to create bookmark')
+				AppGlobal.methods.t('bookmarks', 'Failed to create bookmark'),
 			)
 			throw err
 		}
@@ -298,7 +391,7 @@ export default {
 		try {
 			const response = await axios.put(
 				url(state, `/bookmark/${id}`),
-				this.getters.getBookmark(id)
+				this.getters.getBookmark(id),
 			)
 			const {
 				data: { status },
@@ -312,14 +405,14 @@ export default {
 			commit(mutations.FETCH_END, 'saveBookmark')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to save bookmark')
+				AppGlobal.methods.t('bookmarks', 'Failed to save bookmark'),
 			)
 			throw err
 		}
 	},
 	async [actions.MOVE_BOOKMARK](
 		{ commit, dispatch, state },
-		{ bookmark, oldFolder, newFolder }
+		{ bookmark, oldFolder, newFolder },
 	) {
 		if (Number(oldFolder) === Number(newFolder)) {
 			return
@@ -327,33 +420,35 @@ export default {
 		commit(mutations.FETCH_START, { type: 'moveBookmark' })
 		try {
 			const response = await axios.post(
-				url(state, `/folder/${newFolder}/bookmarks/${bookmark}`)
+				url(state, `/folder/${newFolder}/bookmarks/${bookmark}`),
 			)
 			if (response.data.status !== 'success') {
 				throw new Error(response.data)
 			}
-			const response2 = await axios.delete(
-				url(state, `/folder/${oldFolder}/bookmarks/${bookmark}`)
-			)
-			if (response2.data.status !== 'success') {
-				throw new Error(response2.data)
+			if (oldFolder) {
+				const response2 = await axios.delete(
+					url(state, `/folder/${oldFolder}/bookmarks/${bookmark}`),
+				)
+				if (response2.data.status !== 'success') {
+					throw new Error(response2.data)
+				}
+				dispatch(actions.LOAD_FOLDER_CHILDREN_ORDER, oldFolder)
 			}
 			commit(mutations.FETCH_END, 'moveBookmark')
-			dispatch(actions.LOAD_FOLDER_CHILDREN_ORDER, oldFolder)
 			dispatch(actions.LOAD_FOLDER_CHILDREN_ORDER, newFolder)
 		} catch (err) {
 			console.error(err)
 			commit(mutations.FETCH_END, 'moveBookmark')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to move bookmark')
+				AppGlobal.methods.t('bookmarks', 'Failed to move bookmark'),
 			)
 			throw err
 		}
 	},
 	async [actions.COPY_BOOKMARK](
 		{ commit, dispatch, state },
-		{ bookmark, oldFolder, newFolder }
+		{ bookmark, oldFolder, newFolder },
 	) {
 		if (Number(oldFolder) === Number(newFolder)) {
 			return
@@ -361,7 +456,7 @@ export default {
 		commit(mutations.FETCH_START, { type: 'copyBookmark' })
 		try {
 			const response = await axios.post(
-				url(state, `/folder/${newFolder}/bookmarks/${bookmark}`)
+				url(state, `/folder/${newFolder}/bookmarks/${bookmark}`),
 			)
 			if (response.data.status !== 'success') {
 				throw new Error(response.data)
@@ -373,12 +468,13 @@ export default {
 			commit(mutations.FETCH_END, 'copyBookmark')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to copy bookmark')
+				AppGlobal.methods.t('bookmarks', 'Failed to copy bookmark'),
 			)
 			throw err
 		}
 	},
 	async [actions.CLICK_BOOKMARK]({ commit, dispatch, state }, bookmark) {
+		commit(mutations.CLICK_BOOKMARK, bookmark)
 		commit(mutations.FETCH_START, { type: 'clickBookmark' })
 		try {
 			const response = await axios.post(url(state, '/bookmark/click'), {
@@ -399,12 +495,12 @@ export default {
 	},
 	async [actions.DELETE_BOOKMARK](
 		{ commit, dispatch, state },
-		{ id, folder, avoidReload }
+		{ id, folder, avoidReload, hard },
 	) {
 		if (folder) {
 			try {
 				const response = await axios.delete(
-					url(state, `/folder/${folder}/bookmarks/${id}`)
+					url(state, `/folder/${folder}/bookmarks/${id}` + (hard ? '?hardDelete=true' : '')),
 				)
 				if (response.data.status !== 'success') {
 					throw new Error(response.data)
@@ -420,13 +516,14 @@ export default {
 					mutations.SET_ERROR,
 					AppGlobal.methods.t(
 						'bookmarks',
-						'Failed to delete bookmark'
-					)
+						'Failed to delete bookmark',
+					),
 				)
 				throw err
 			}
 			return
 		}
+
 		try {
 			const response = await axios.delete(url(state, `/bookmark/${id}`))
 			if (response.data.status !== 'success') {
@@ -438,14 +535,46 @@ export default {
 			console.error(err)
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to delete bookmark')
+				AppGlobal.methods.t('bookmarks', 'Failed to delete bookmark'),
+			)
+			throw err
+		}
+	},
+	async [actions.UNDELETE_BOOKMARK](
+		{ commit, dispatch, state, getters },
+		{ id, folder, avoidReload },
+	) {
+		try {
+			const response = await axios.post(
+				url(state, `/folder/${folder}/bookmarks/${id}/undelete`),
+			)
+			if (response.data.status !== 'success') {
+				throw new Error(response.data)
+			}
+			commit(mutations.REMOVE_BOOKMARK, id)
+			if (!avoidReload) {
+				await dispatch(actions.COUNT_BOOKMARKS, -1)
+				await dispatch(actions.LOAD_FOLDER_CHILDREN_ORDER, folder)
+			}
+			const folderItem = getters.getFolder(folder)[0]
+			if (folderItem.softDeleted) {
+				await dispatch(actions.MOVE_BOOKMARK, { bookmark: id, newFolder: -1 })
+			}
+		} catch (err) {
+			console.error(err)
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t(
+					'bookmarks',
+					'Failed to restore bookmark',
+				),
 			)
 			throw err
 		}
 	},
 	async [actions.IMPORT_BOOKMARKS](
 		{ commit, dispatch, state },
-		{ file, folder }
+		{ file, folder },
 	) {
 		commit(mutations.FETCH_START, { type: 'importBookmarks' })
 		const data = new FormData()
@@ -453,7 +582,7 @@ export default {
 		try {
 			const response = await axios.post(
 				url(state, `/folder/${folder || -1}/import`),
-				data
+				data,
 			)
 			if (!response.data || response.data.status !== 'success') {
 				if (response.status === 413) {
@@ -463,7 +592,7 @@ export default {
 				throw new Error(
 					Array.isArray(response.data.data)
 						? response.data.data.join('. ')
-						: response.data.data
+						: response.data.data,
 				)
 			}
 			await dispatch(actions.COUNT_BOOKMARKS, -1)
@@ -472,7 +601,7 @@ export default {
 			commit(mutations.FETCH_END, 'importBookmarks')
 			return commit(
 				mutations.SET_NOTIFICATION,
-				AppGlobal.methods.t('bookmarks', 'Import successful')
+				AppGlobal.methods.t('bookmarks', 'Import successful'),
 			)
 		} catch (err) {
 			console.error(err)
@@ -502,7 +631,7 @@ export default {
 				commit(mutations.FETCH_END, 'deleteBookmarks')
 				commit(
 					mutations.SET_ERROR,
-					AppGlobal.methods.t('bookmarks', err.message)
+					AppGlobal.methods.t('bookmarks', err.message),
 				)
 				throw err
 			})
@@ -510,7 +639,7 @@ export default {
 
 	async [actions.RENAME_TAG](
 		{ commit, dispatch, state },
-		{ oldName, newName }
+		{ oldName, newName },
 	) {
 		commit(mutations.FETCH_START, { type: 'tag' })
 		try {
@@ -531,12 +660,23 @@ export default {
 			commit(mutations.FETCH_END, 'tag')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to rename tag')
+				AppGlobal.methods.t('bookmarks', 'Failed to rename tag'),
 			)
 			throw err
 		}
 	},
 	[actions.LOAD_TAGS]({ commit, dispatch, state }) {
+		if (state.tags === null) {
+			try {
+				const tags = loadState('bookmarks', 'tags')
+				return commit(mutations.SET_TAGS, tags)
+			} catch (e) {
+				console.warn(e)
+				console.warn(
+					'Could not load initial bookmarks count state, continuing with HTTP request',
+				)
+			}
+		}
 		commit(mutations.FETCH_START, { type: 'tags' })
 		return axios
 			.get(url(state, '/tag'), { params: { count: true } })
@@ -550,7 +690,7 @@ export default {
 				commit(mutations.FETCH_END, 'tags')
 				commit(
 					mutations.SET_ERROR,
-					AppGlobal.methods.t('bookmarks', 'Failed to load tags')
+					AppGlobal.methods.t('bookmarks', 'Failed to load tags'),
 				)
 				throw err
 			})
@@ -574,8 +714,8 @@ export default {
 					mutations.SET_ERROR,
 					AppGlobal.methods.t(
 						'bookmarks',
-						'Failed to delete bookmark'
-					)
+						'Failed to delete bookmark',
+					),
 				)
 				throw err
 			})
@@ -588,7 +728,7 @@ export default {
 				return commit(mutations.SET_FOLDERS, folders)
 			} catch (e) {
 				console.warn(
-					'Could not load initial folder state, continuing with HTTP request'
+					'Could not load initial folder state, continuing with HTTP request',
 				)
 			}
 		}
@@ -616,17 +756,94 @@ export default {
 			commit(mutations.FETCH_END, 'folders')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to load folders')
+				AppGlobal.methods.t('bookmarks', 'Failed to load folders'),
 			)
 			throw err
 		}
 	},
+
+	async [actions.LOAD_DELETED_FOLDERS]({ commit, dispatch, state }) {
+		if (state.deletedFolders === null) {
+			try {
+				const count = loadState('bookmarks', 'deletedFolders')
+				return commit(mutations.SET_DELETED_FOLDERS, count)
+			} catch (e) {
+				console.warn(
+					'Could not load initial deleted folders state, continuing with HTTP request',
+				)
+			}
+		}
+		let canceled = false
+		commit(mutations.FETCH_START, {
+			type: 'deleted_folders',
+			cancel: () => {
+				canceled = true
+			},
+		})
+		try {
+			const response = await axios.get(url(state, '/folder/deleted'), {
+				params: {},
+			})
+			if (canceled) return
+			const {
+				data: { data, status },
+			} = response
+			if (status !== 'success') throw new Error(data)
+			const folders = data
+			commit(mutations.FETCH_END, 'folders')
+			return commit(mutations.SET_DELETED_FOLDERS, folders)
+		} catch (err) {
+			console.error(err)
+			commit(mutations.FETCH_END, 'folders')
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t('bookmarks', 'Failed to load deleted folders'),
+			)
+			throw err
+		}
+	},
+
+	async [actions.LOAD_DELETED_BOOKMARKS]({ commit, dispatch, state }) {
+		let canceled = false
+		commit(mutations.FETCH_START, {
+			type: 'bookmarks',
+			cancel: () => {
+				canceled = true
+			},
+		})
+		try {
+			const response = await axios.get(url(state, '/bookmark/deleted'), {
+				params: {},
+			})
+			if (canceled) return
+			const {
+				data: { data, status },
+			} = response
+			if (status !== 'success') throw new Error(data)
+			const bookmarks = data
+			commit(mutations.FETCH_END, 'bookmarks')
+			commit(mutations.REMOVE_ALL_BOOKMARKS)
+			for (const bookmark of bookmarks) {
+				commit(mutations.ADD_BOOKMARK, bookmark)
+			}
+			commit(mutations.REACHED_END)
+		} catch (err) {
+			console.error(err)
+			commit(mutations.FETCH_END, 'bookmarks')
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t('bookmarks', 'Failed to load deleted bookmarks'),
+			)
+			throw err
+		}
+	},
+
 	async [actions.DELETE_FOLDER](
 		{ commit, dispatch, state },
-		{ id, avoidReload }
+		{ id, avoidReload, hard },
 	) {
 		try {
-			const response = await axios.delete(url(state, `/folder/${id}`))
+			const response = await axios.delete(url(state, `/folder/${id}` + (hard ? '?hardDelete=true' : '')))
 			const {
 				data: { status },
 			} = response
@@ -637,22 +854,86 @@ export default {
 			if (!avoidReload) {
 				await dispatch(
 					actions.LOAD_FOLDER_CHILDREN_ORDER,
-					parentFolder
+					parentFolder,
 				)
 				await dispatch(actions.LOAD_FOLDERS)
+				await dispatch(actions.LOAD_DELETED_FOLDERS)
 			}
 		} catch (err) {
 			console.error(err)
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to delete folder')
+				AppGlobal.methods.t('bookmarks', 'Failed to delete folder'),
+			)
+			throw err
+		}
+	},
+	async [actions.MOVE_FOLDER](
+		{ commit, dispatch, state },
+		{ folderId, targetFolderId },
+	) {
+		try {
+			const folder = this.getters.getFolder(folderId)[0]
+			commit(mutations.MOVE_FOLDER, { folder: folderId, target: targetFolderId })
+			const oldParent = folder.parent_folder
+			folder.parent_folder = targetFolderId
+			try {
+				await dispatch(actions.SAVE_FOLDER, folder.id) // reloads children order for new parent
+				dispatch(
+					actions.LOAD_FOLDER_CHILDREN_ORDER,
+					oldParent,
+				)
+			} catch (err) {
+				commit(mutations.MOVE_FOLDER, { folder: folder.id, target: oldParent })
+				folder.parent_folder = oldParent
+				throw err
+			}
+		} catch (err) {
+			console.error(err)
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t('bookmarks', 'Failed to move folder'),
+			)
+			throw err
+		}
+	},
+	async [actions.UNDELETE_FOLDER](
+		{ commit, dispatch, state },
+		{ id, avoidReload },
+	) {
+		try {
+			const parentFolderId = this.getters.getFolder(id)[0].parent_folder
+			const parentFolderItem = this.getters.getFolder(parentFolderId)[0]
+			const response = await axios.post(url(state, `/folder/${id}/undelete`))
+			const {
+				data: { status },
+			} = response
+			if (status !== 'success') {
+				throw new Error(response.data)
+			}
+			if (parentFolderItem && parentFolderItem.softDeleted) {
+				await dispatch(actions.MOVE_FOLDER, { folderId: id, targetFolderId: '-1' })
+			}
+			if (!avoidReload) {
+				await dispatch(
+					actions.LOAD_FOLDER_CHILDREN_ORDER,
+					parentFolderId,
+				)
+				await dispatch(actions.LOAD_FOLDERS)
+				await dispatch(actions.LOAD_DELETED_FOLDERS)
+			}
+		} catch (err) {
+			console.error(err)
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t('bookmarks', 'Failed to restore folder'),
 			)
 			throw err
 		}
 	},
 	async [actions.CREATE_FOLDER](
 		{ commit, dispatch, state },
-		{ parentFolder, title }
+		{ parentFolder, title },
 	) {
 		try {
 			const response = await axios.post(url(state, '/folder'), {
@@ -668,14 +949,14 @@ export default {
 			commit(mutations.DISPLAY_NEW_FOLDER, false)
 			await dispatch(
 				actions.LOAD_FOLDER_CHILDREN_ORDER,
-				parentFolder || -1
+				parentFolder || -1,
 			)
 			await dispatch(actions.LOAD_FOLDERS, /* force: */ true)
 		} catch (err) {
 			console.error(err)
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to create folder')
+				AppGlobal.methods.t('bookmarks', 'Failed to create folder'),
 			)
 			throw err
 		}
@@ -697,7 +978,7 @@ export default {
 			}
 			await dispatch(
 				actions.LOAD_FOLDER_CHILDREN_ORDER,
-				folder.parent_folder
+				folder.parent_folder,
 			)
 			commit(mutations.FETCH_END, 'saveFolder')
 		} catch (err) {
@@ -705,19 +986,19 @@ export default {
 			commit(mutations.FETCH_END, 'saveFolder')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.methods.t('bookmarks', 'Failed to create folder')
+				AppGlobal.methods.t('bookmarks', 'Failed to create folder'),
 			)
 			throw err
 		}
 	},
 	async [actions.LOAD_FOLDER_CHILDREN_ORDER](
 		{ commit, dispatch, state },
-		id
+		id,
 	) {
 		commit(mutations.FETCH_START, { type: 'childrenOrder' })
 		try {
 			const response = await axios.get(
-				url(state, `/folder/${id}/childorder`)
+				url(state, `/folder/${id}/childorder`),
 			)
 			const {
 				data: { status },
@@ -737,8 +1018,8 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to load children order'
-				)
+					'Failed to load children order',
+				),
 			)
 			throw err
 		}
@@ -756,49 +1037,38 @@ export default {
 			await Parallel.each(
 				state.selection.folders,
 				async folder => {
-					if (folderId === folder.id) {
-						throw new Error('Cannot move folder into itself')
-					}
-					const oldParent = folder.parent_folder
-					folder.parent_folder = folderId
-					await dispatch(actions.SAVE_FOLDER, folder.id) // reloads children order for new parent
-					await dispatch(
-						actions.LOAD_FOLDER_CHILDREN_ORDER,
-						oldParent
-					)
+					await dispatch(actions.MOVE_FOLDER, { folderId: folder.id, targetFolderId: folderId })
 				},
-				10
+				10,
+			)
+			await Parallel.each(
+				state.selection.bookmarks,
+				bookmark => {
+					commit(mutations.REMOVE_BOOKMARK, bookmark.id)
+					return dispatch(actions.MOVE_BOOKMARK, {
+						oldFolder:
+						bookmark.folders[bookmark.folders.length - 1], // FIXME This is veeeery ugly and will cause issues. Inevitably.
+						newFolder: folderId,
+						bookmark: bookmark.id,
+					})
+				},
+				10,
 			)
-			await Promise.all([
-				dispatch(actions.LOAD_FOLDERS),
-				Parallel.each(
-					state.selection.bookmarks,
-					bookmark => {
-						commit(mutations.REMOVE_BOOKMARK, bookmark.id)
-						return dispatch(actions.MOVE_BOOKMARK, {
-							oldFolder:
-							bookmark.folders[bookmark.folders.length - 1], // FIXME This is veeeery ugly and will cause issues. Inevitably.
-							newFolder: folderId,
-							bookmark: bookmark.id,
-						})
-					},
-					10
-				),
-			])
 
 			// Because we're possibly moving across share boundaries we need to recount
-			await dispatch(actions.COUNT_BOOKMARKS, -1)
+			dispatch(actions.COUNT_BOOKMARKS, -1)
 
 			commit(mutations.FETCH_END, 'moveSelection')
 		} catch (err) {
 			console.error(err)
+			console.error(err.list)
 			commit(mutations.FETCH_END, 'moveSelection')
 			commit(
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to move parts of selection'
-				)
+					'Failed to move parts of selection',
+				),
 			)
 			throw err
 		}
@@ -809,31 +1079,23 @@ export default {
 			await Parallel.each(
 				state.selection.folders,
 				async folder => {
-					if (folderId === folder.id) {
-						throw new Error('Cannot copy folder into itself')
+					if (folder) {
+						throw new Error('Cannot copy folders')
 					}
-					const oldParent = folder.parent_folder
-					folder.parent_folder = folderId
-					await dispatch(actions.SAVE_FOLDER, folder.id) // reloads children order for new parent
-					await dispatch(
-						actions.LOAD_FOLDER_CHILDREN_ORDER,
-						oldParent
-					)
 				},
-				10
+				10,
 			)
 			await Promise.all([
 				dispatch(actions.LOAD_FOLDERS),
 				Parallel.each(
 					state.selection.bookmarks,
 					bookmark => {
-						commit(mutations.REMOVE_BOOKMARK, bookmark.id)
 						return dispatch(actions.COPY_BOOKMARK, {
 							newFolder: folderId,
 							bookmark: bookmark.id,
 						})
 					},
-					10
+					10,
 				),
 			])
 
@@ -845,8 +1107,8 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to copy parts of selection'
-				)
+					'Failed to copy parts of selection',
+				),
 			)
 			throw err
 		}
@@ -861,17 +1123,29 @@ export default {
 						id: folder.id,
 						avoidReload: true,
 					}),
-				10
+				10,
 			)
 			await Parallel.each(
 				state.selection.bookmarks,
-				bookmark =>
-					dispatch(actions.DELETE_BOOKMARK, {
-						id: bookmark.id,
-						folder,
-						avoidReload: true,
-					}),
-				10
+				(bookmark) => {
+					if (folder) {
+						return dispatch(actions.DELETE_BOOKMARK, {
+							id: bookmark.id,
+							folder,
+							avoidReload: true,
+						})
+					} else {
+						// soft delete all occurences instead of hard deleting the bookmark itself
+						return Promise.all(bookmark.folders.map((folder) => {
+							return dispatch(actions.DELETE_BOOKMARK, {
+								id: bookmark.id,
+								folder,
+								avoidReload: true,
+							})
+						}))
+					}
+				},
+				10,
 			)
 			dispatch(actions.RELOAD_VIEW)
 			commit(mutations.FETCH_END, 'deleteSelection')
@@ -882,8 +1156,8 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to delete parts of selection'
-				)
+					'Failed to delete parts of selection',
+				),
 			)
 			throw err
 		}
@@ -902,7 +1176,7 @@ export default {
 						return dispatch(actions.SAVE_BOOKMARK, bookmark.id)
 					}
 				},
-				10
+				10,
 			)
 			await dispatch(actions.LOAD_TAGS)
 			commit(mutations.FETCH_END, 'tagSelection')
@@ -913,8 +1187,8 @@ export default {
 				mutations.SET_ERROR,
 				AppGlobal.methods.t(
 					'bookmarks',
-					'Failed to tag parts of selection'
-				)
+					'Failed to tag parts of selection',
+				),
 			)
 			throw err
 		}
@@ -942,8 +1216,8 @@ export default {
 		commit(mutations.SET_QUERY, { sortby: 'clickcount' })
 		return dispatch(actions.FETCH_PAGE)
 	},
-	[actions.FILTER_BY_SEARCH]({ dispatch, commit }, search) {
-		commit(mutations.SET_QUERY, { search: search.split(' ') })
+	[actions.FILTER_BY_SEARCH]({ dispatch, commit }, { search, folder }) {
+		commit(mutations.SET_QUERY, { search: search.split(' '), conjunction: 'and', ...(Number(folder) !== -1 && { folder, recursive: true }) })
 		return dispatch(actions.FETCH_PAGE)
 	},
 	[actions.FILTER_BY_TAGS]({ dispatch, commit }, tags) {
@@ -962,26 +1236,30 @@ export default {
 		commit(mutations.SET_QUERY, { archived: true })
 		return dispatch(actions.FETCH_PAGE)
 	},
-	[actions.FILTER_BY_FOLDER]({ dispatch, commit, state }, folder) {
-		commit(mutations.SET_QUERY, { folder })
+	[actions.FILTER_BY_DUPLICATED]({ dispatch, commit }) {
+		commit(mutations.SET_QUERY, { duplicated: true })
+		return dispatch(actions.FETCH_PAGE)
+	},
+	[actions.FILTER_BY_FOLDER]({ dispatch, commit, state }, { folder, softDeleted }) {
+		commit(mutations.SET_QUERY, { folder, deleted: softDeleted })
 		if (state.settings.sorting === 'index') {
 			dispatch(actions.LOAD_FOLDER_CHILDREN_ORDER, folder)
 		}
 		return dispatch(actions.FETCH_PAGE)
 	},
 
-	[actions.FETCH_PAGE]({ dispatch, commit, state }) {
+	async [actions.FETCH_PAGE]({ dispatch, commit, state }) {
 		if (state.fetchState.reachedEnd) return
 		if (state.loading.bookmarks) return
 		let canceled = false
 		const fetchedPage = state.fetchState.page
-		commit(mutations.FETCH_START, {
+		await commit(mutations.FETCH_START, {
 			type: 'bookmarks',
 			cancel() {
 				canceled = true
 			},
 		})
-		axios
+		return axios
 			.get(url(state, '/bookmark'), {
 				params: {
 					limit: BATCH_SIZE,
@@ -1012,7 +1290,7 @@ export default {
 				commit(mutations.FETCH_END, 'bookmarks')
 				commit(
 					mutations.SET_ERROR,
-					AppGlobal.t('bookmarks', 'Failed to fetch bookmarks.')
+					AppGlobal.t('bookmarks', 'Failed to fetch bookmarks.'),
 				)
 				throw err
 			})
@@ -1049,7 +1327,7 @@ export default {
 			commit(mutations.FETCH_END, 'bookmarks')
 			commit(
 				mutations.SET_ERROR,
-				AppGlobal.t('bookmarks', 'Failed to fetch bookmarks.')
+				AppGlobal.t('bookmarks', 'Failed to fetch bookmarks.'),
 			)
 			throw err
 		}
@@ -1068,13 +1346,13 @@ export default {
 		}
 		return axios
 			.post(url(state, `/settings/${key}`), {
-				[key]: value,
+				value: String(value),
 			})
 			.catch(err => {
 				console.error(err)
 				commit(
 					mutations.SET_ERROR,
-					AppGlobal.methods.t('bookmarks', 'Failed to change setting')
+					AppGlobal.methods.t('bookmarks', 'Failed to store setting'),
 				)
 				throw err
 			})
@@ -1103,20 +1381,69 @@ export default {
 					AppGlobal.methods.t(
 						'bookmarks',
 						'Failed to load setting {key}',
-						{ key }
-					)
+						{ key },
+					),
 				)
 				throw err
 			})
 	},
-	[actions.LOAD_SETTINGS]({ commit, dispatch, state }) {
-		return Promise.all(
-			['sorting', 'viewMode', 'archivePath', 'limit'].map(key =>
-				dispatch(actions.LOAD_SETTING, key)
-			)
-		)
+	async [actions.LOAD_SETTINGS]({ commit, dispatch, state }) {
+		const settings = loadState('bookmarks', 'settings')
+		for (const setting in settings) {
+			const key = setting
+			let value = settings[setting]
+			switch (key) {
+			case 'viewMode':
+				value = value || state.settings.viewMode
+				await commit(mutations.SET_VIEW_MODE, value)
+				break
+			case 'sorting':
+				value = value || state.settings.sorting
+				await commit(mutations.RESET_PAGE)
+				break
+			}
+			await commit(mutations.SET_SETTING, { key, value })
+		}
 	},
 
+	[actions.LOAD_SHARES]({ commit, dispatch, state }) {
+		return axios
+			.get(url(state, '/share'))
+			.then(async response => {
+				const {
+					data: { data, status },
+				} = response
+				if (status !== 'success') throw new Error(data)
+				const shares = data
+				for (const share of shares) {
+					await commit(mutations.ADD_SHARE, share)
+				}
+			})
+			.catch(err => {
+				console.error(err)
+				// Don't set a notification as this is expected to happen for subfolders of shares that we don't have a RESHAR permission for
+				throw err
+			})
+	},
+	[actions.LOAD_SHARED_FOLDERS]({ commit, dispatch, state }) {
+		return axios
+			.get(url(state, '/folder/shared'))
+			.then(async response => {
+				const {
+					data: { data, status },
+				} = response
+				if (status !== 'success') throw new Error(data)
+				const folders = data
+				for (const folder of folders) {
+					await commit(mutations.ADD_SHARED_FOLDER, folder)
+				}
+			})
+			.catch(err => {
+				console.error(err)
+				// Don't set a notification as this is expected to happen for subfolders of shares that we don't have a RESHAR permission for
+				throw err
+			})
+	},
 	[actions.LOAD_SHARES_OF_FOLDER]({ commit, dispatch, state }, folderId) {
 		if (folderId === -1 || folderId === '-1') {
 			return Promise.resolve()
@@ -1141,7 +1468,7 @@ export default {
 	},
 	[actions.CREATE_SHARE](
 		{ commit, dispatch, state },
-		{ folderId, type, participant }
+		{ folderId, type, participant },
 	) {
 		return axios
 			.post(url(state, `/folder/${folderId}/shares`), {
@@ -1163,15 +1490,15 @@ export default {
 					AppGlobal.methods.t(
 						'bookmarks',
 						'Failed to create share for folder {folderId}',
-						{ folderId }
-					)
+						{ folderId },
+					),
 				)
 				throw err
 			})
 	},
 	[actions.EDIT_SHARE](
 		{ commit, dispatch, state },
-		{ shareId, canWrite, canShare }
+		{ shareId, canWrite, canShare },
 	) {
 		return axios
 			.put(url(state, `/share/${shareId}`), {
@@ -1192,8 +1519,8 @@ export default {
 					AppGlobal.methods.t(
 						'bookmarks',
 						'Failed to update share {shareId}',
-						{ shareId }
-					)
+						{ shareId },
+					),
 				)
 				throw err
 			})
@@ -1215,8 +1542,8 @@ export default {
 					AppGlobal.methods.t(
 						'bookmarks',
 						'Failed to delete share {shareId}',
-						{ shareId }
-					)
+						{ shareId },
+					),
 				)
 				throw err
 			})
@@ -1261,8 +1588,8 @@ export default {
 					AppGlobal.methods.t(
 						'bookmarks',
 						'Failed to create public link for folder {folderId}',
-						{ folderId }
-					)
+						{ folderId },
+					),
 				)
 				throw err
 			})
@@ -1284,12 +1611,59 @@ export default {
 					AppGlobal.methods.t(
 						'bookmarks',
 						'Failed to delete public link for folder {folderId}',
-						{ folderId }
-					)
+						{ folderId },
+					),
 				)
 				throw err
 			})
 	},
+	async [actions.EMPTY_TRASHBIN]({ commit, dispatch, state }) {
+		if (state.loading.emptyTrashbin) return
+		await commit(mutations.FETCH_START, {
+			type: 'emptyTrashbin',
+		})
+		try {
+			await Parallel.each(
+				state.deletedFolders,
+				folder =>
+					dispatch(actions.DELETE_FOLDER, {
+						id: folder.id,
+						avoidReload: true,
+						hard: true,
+					}),
+				10,
+			)
+			await Parallel.each(
+				state.bookmarks,
+				(bookmark) => {
+					// soft delete all occurences instead of hard deleting the bookmark itself
+					return Promise.all(bookmark.folders.map((folder) => {
+						return dispatch(actions.DELETE_BOOKMARK, {
+							id: bookmark.id,
+							folder,
+							avoidReload: true,
+							hard: true,
+						})
+					}))
+				},
+				10,
+			)
+			dispatch(actions.RELOAD_VIEW)
+			dispatch(actions.LOAD_DELETED_FOLDERS)
+			commit(mutations.FETCH_END, 'emptyTrashbin')
+		} catch (err) {
+			console.error(err)
+			commit(mutations.FETCH_END, 'emptyTrashbin')
+			commit(
+				mutations.SET_ERROR,
+				AppGlobal.methods.t(
+					'bookmarks',
+					'Failed to permanently delete parts of trash bin',
+				),
+			)
+			throw err
+		}
+	},
 }
 
 /**
